unit XuDsigV;
{This is the Unit that connects to the MicroSoft CryptoAPI
 that is used to create a digital Signature for a block of
 Data.
}
interface

uses SysUtils, Windows, WCrypt2, XlfMime, Dialogs, classes,
     XuDsigConst;

type
    ECryptErr = Exception;
                                                                                                  //This function is to check the all the input data is ready

type
    tCryptography = class(tobject)
    private
       //Basic Crypto data
       hProv: HCRYPTPROV; // provider handle
       hPassKey: HCRYPTKEY; // Handle to password (key)
       hHash: HCRYPTHASH; //Hash handle
       hUserKey: HCRYPTKEY;  //Handle to the users Sig key
       hPubKey: HCRYPTKEY;  //Public Key Handle
       //Cert. Store Data
       hCertStore, hCRLstore: HCERTSTORE;  //Certificate store handle
       hMsg: HCRYPTMSG;
       pCertContext: PCCERT_CONTEXT;  //Cert Context

       //Local data
       fProviderName, fContainerName, fCryptVer: string; // Misc. info
       fHexHash: string;  //hash value in Hex
       fHashAlogrithm: string;  //Alogrithm used
       fDataBuffer: string;  //This is the data to sign as one big string
       fSignatureStr: string;   //This is the signature passed in.
       fDateSigned:  string;  //FM datetime
       datesigned: FILETIME;  //The date signed converted to system filetime.
       SignMsgParam: CRYPT_SIGN_MESSAGE_PARA;
       LastErr: DWORD;  //Error status from calls
       procedure RaiseErr(msg: string);
       function GetPropertyFromKPI(buff: array of char; prop: integer): string;
       function UnicodeToStr(pba: pByteArray; var offSet: integer): string;
       procedure FixSigTime(MDT: string; var cryptoDT: filetime);
       function DataReady: boolean;
       procedure CRLOpenStore;
       procedure CRLCloseStore;
       function CertChainCheck(pCertContext: PCCERT_CONTEXT): boolean;
       function CertRevocation(pCertContext: PCCERT_CONTEXT): boolean;
    public
       //These fields hold the raw data from the Crypto functions
       CHashValue: Array of byte;
       CKeyBlob, CSignature: Array[0..2024] of byte;
       CHashLen, CBlobLen, CSigLen: integer;  //The length
       HashB64, KeyBloB64: string;
       CertName, Certstr: string; //pass back data
       ErrCode, ReturnStr: string;  //
       TrackingMsg: Tstringlist;

       constructor Create;
       destructor destroy; override;
       //Property's
       property ProviderName: string read fProviderName;
       property ContainerName: string read fContainerName;
       property CryptVer: string read fCryptVer;
       property HashAlgorithm: string read fHashAlogrithm;
       property DTSigned: string read fDateSigned write fDateSigned;
       property HexHash: string read fHexHash;
       property DataBuffer: string read fDataBuffer write fDataBuffer;
       property SignatureStr: string read fSignatureStr write fSignatureStr;
       //Functions and Procedures
       procedure HashStart;
       procedure HashBuf(buffer: Pbyte; Count: integer);
       procedure GetHashValue;
       procedure HashEnd;
       function VerifySigb64(keyb64, sigb64: string): bool;
       procedure B64ToArray(S: string; var buf: array of byte; var len: integer);
       procedure CertOpen(c: integer);
       procedure CertClose;
       function VerifyData: boolean;
       function CheckCertAgenstCRL(v:pointer): boolean; //(pcrtinfo: PCERT_INFO):boolean;

    end;



implementation


procedure tCryptography.RaiseErr(msg: string);
// Raise an exception with a formatted message
var
   err: dword;
   s, ec: string;
begin
    s := '';
    ErrCode := msg; //Put in public variable
    err := getLastError; // Check last system error
    if '8980' = copy(msg, 1, 4) then //Is it a local error.
       raise eCryptErr.Create(msg);
       
    if SUCCEEDED(err) then
       raise eCryptErr.createFmt('%s^Contact IRM: Crypto error: %s.', [msg]);

    s := SysErrorMessage(err);
    if s = '' then
       s := format('System Error %0X', [err]);
    ErrCode := format('89802010^Contact IRM: Crypto error: %s (%s).', [msg, s]);
    raise eCryptErr.createFmt('Cryptography error: %s (%s).', [msg, s]);
end;

//This function is to check the all the input data is ready
function tCryptography.DataReady: boolean;
begin
     Result := false;
     if length(fDataBuffer) = 0 then exit;
     Result := true;
end;


//This will create and init the hprov object
constructor tCryptography.Create;

var
   buff: array [0..1023] of char;
   provider, Container: string;
   canCreate: boolean;
   size: dWord;
   LastErr: DWORD;

begin
   inherited Create;
   hProv := 0;      //See that global variables are inited
   hHash := 0;
   hCertStore := 0;
   fProviderName := '';
   fContainerName := '';
   fCryptVer := '0.0';
   CHashLen := 0;
   CSigLen := 0;
   CBlobLen := 0;
   fDataBuffer := '';
   fSignatureStr := '';
   fDatesigned := '';
   //
   Trackingmsg := TstringList.Create;

   //Find out what CSP to use.
   //For Verification use the MS enhansed AES version.
     provider := cVPROV_NAME;    //Changed from '', rwf
     Container := 'MY';  //Use the default container 'MY'
     canCreate := true;
     // Create a cryptography object with default
     // container and slected provider.


   if CryptAcquireContext( @hProv,
                           pchar(Container),
                           pchar(Provider),
                           PROV_RSA_AES,  // was PROV_RSA_FULL,   rwf
                           0) = false then
      // Unable to aquire context, check if failure was due to
      // absence of key container.
      if (CanCreate) and (GetLastError = NTE_BAD_KEYSET) then
         // Create a new key set for the current user.
         if CryptAcquireContext( @hProv,
                                 pchar(Container),
                                 pchar(Provider),
                                 PROV_RSA_AES,  // was PROV_RSA_FULL,  rwf
                                 CRYPT_NEWKEYSET) = false then
            RaiseErr( 'Unable to aquire context');

      lastErr := GetLastError;
      if lastErr = NTE_BAD_SIGNATURE then RaiseErr( 'Bad DLL ');
      //We get the next error if the SmartCard is not in the Reader
      if lastErr = NTE_FAIL then RaiseErr('SmartCard NOT in Reader');
      if lastErr = NTE_KEYSET_ENTRY_BAD then RaiseErr('Err');
      if lastErr = NTE_KEYSET_NOT_DEF then RaiseErr('No Container');

   // Get some info about the cryptography module.
   // Name of the CSP.
   size := sizeof(buff);
   if CryptGetProvParam(hProv, PP_NAME, @Buff, @size, 0) = false then
      fProviderName := 'Unknown'
   else
      fProviderName := StrPas(Buff);

   // Name of the key container.
   size := sizeof(buff);
   if CryptGetProvParam(hProv, PP_CONTAINER, @Buff, @size, 0) = false then
      fContainerName := 'Unknown'
   else
      fContainerName := StrPas(Buff);

   // Version number
   size := sizeof(buff);
   if CryptGetProvParam(hProv, PP_VERSION, @Buff, @size, 0) = false then
      fCryptVer := '*.00'
   else
      fCryptVer := format('%d.%d', [integer(buff[1]), integer(buff[0])]);

end;


destructor tCryptography.Destroy;
begin
   try
      // Release the key handle
      if hPassKey <> 0 then
         if CryptDestroyKey(hPassKey) = false then
            RaiseErr('Unable to destroy password key');
      // Release the Hash handle
      if hHash <> 0 then
              CryptDestroyHash(hHash);
      //Release a certificate Context
      if pCertContext <> nil then
              CertFreeCertificateContext(pCertContext);
      //Release the Cert Store.
      if hCertStore <> 0 then
         CertCloseStore(hCertStore, CERT_CLOSE_STORE_FORCE_FLAG);
             // Clean up.
      // Release the CSP handle
      if hProv <> 0 then
         if CryptReleaseContext(hProv, 0) = false then
            RaiseErr('Unable to release context');

   finally
   //Must remember to do this
   inherited;
   end;

end;


//Called at start of a long hash to create the hash object.
procedure tCryptography.Hashstart;
begin
     HashB64 := '';
     CHashLen := 0;
     if hHash = 0 then
        if cryptCreateHash(hprov,cHASH_ALGID,0,0,@hHash) = false
           then RaiseErr('Unable to create hash object');

end;

//Called to hash a buffer of data
procedure tCryptography.HashBuf(buffer: Pbyte; Count: integer);
begin
     if CryptHashData(hHash, buffer, Count, 0) = false
        then RaiseErr('Unable to hash buff');
end;


//This is called at the end of a long hash to destroy the hash object
//If going to sign the hash it must be done before calling this.
procedure tCryptography.Hashend;
begin
     //Destroy the Hash
     CryptDestroyHash(hHash);
     hHash := 0;

end;

//This call gets the current Hash value and puts it in the
//public variable  cHashValue.

procedure tCryptography.GetHashValue;
var
   pbData: pbyte;
   dwLen, LastErr: DWORD;
   i: integer;
begin
     CHashLen := 0;
     if hHash = 0 then exit;
     //Get the length of the hash
     CryptGetHashParam(hHash, HP_HASHVAL, nil, @dwLen, 0);

     GetMem(pbData, dwLen);
     if CryptGetHashParam(hHash, HP_HASHVAL, pbData, @dwLen, 0) <> true then
        begin
           lastErr := GetLastError;
           if lastErr = ERROR_INVALID_HANDLE then
              RaiseErr('Invalid Handle');
           if lastErr = ERROR_INVALID_PARAMETER then
              RaiseErr('Invalid Parameter');
           if lastErr = ERROR_MORE_DATA then
              RaiseErr('Error More Data - buffer to small')
           else RaiseErr('Error code' + inttostr(lastErr));
           end;
     CHashLen := dwLen;
     SetLength(cHashValue, dwLen);

     //Convert the hash to a Hex string
     for i := 0 to dwLen-1 do
          cHashValue[i] := pbytearray(pbData)[i];


end;



//This function takes a Public key and signature both in base64 and
//does the VerifySignature with the current Hash.
function tCryptography.VerifySigb64(keyb64, sigb64: string): bool;
var
   i, lastErr: integer;
   pbSignature: pbyte;
begin
     //Load the Public Key
     //SetPublicKey(keyb64);
     B64ToArray(sigb64, CSignature, CsigLen);
     pbSignature := @CSignature;
     Result := CryptVerifySignature(hHash, pbSignature, DWORD(CsigLen),hPubKey, Pchar(''), 0);
     if not Result then
           begin
           lastErr := GetLastError;
           if lastErr = ERROR_INVALID_HANDLE then
              ErrCode :='Invalid Handle';
           if lastErr = ERROR_INVALID_PARAMETER then
              ErrCode := 'Invalid Parameter';
           if lastErr = ERROR_MORE_DATA then
              ErrCode := 'Error More Data - buffer to small'
           //else ErrCode := 'Error code' + inttostr(lastErr);
           else ErrCode := 'Error code' + SysErrorMessage(lastErr);
           end;      //s := SysErrorMessage(err);
end;

//Open a Cert store
// 1=Memory, 2=System, 3=FileName
procedure Tcryptography.CertOpen(c: integer);
var
   prov: LPCSTR;  //What cert store to use
   flag, lasterr: DWORD;  //Cert store control
   ENCODING_TYPE: DWORD; //
   pvPara: PVOID; //Pointer to a string or nil
   filename,n: string;
begin
    hCertStore := 0;
    prov := CERT_STORE_PROV_MEMORY;
    pvPara := nil;
case c of
  1: begin  //Open a memory store
    prov := CERT_STORE_PROV_MEMORY;
    flag := CERT_SYSTEM_STORE_CURRENT_USER;
    hCertStore := CertOpenStore(prov,
                    0,
                    0,
                    flag,
                    pvPara);

     end;

  2: begin //Open a system store
       prov := CERT_STORE_PROV_SYSTEM;
       flag := CERT_SYSTEM_STORE_CURRENT_USER;
       n := 'UserDS'; //'MY';
       pvPara := @n;
       hCertStore := CertOpenStore(prov,
                    0,
                    0,
                    flag,
                    pvPara);
     end;
  3: begin  //Open a Filename store
       n := 'FORTW';
       prov := CERT_STORE_PROV_FILENAME_A;  //Ascii name
       flag := CERT_SYSTEM_STORE_CURRENT_USER;
       filename := 'C:\PROJECTS\CRYPTOAPI\iveysmime.p7s';
       ENCODING_TYPE := PKCS_7_ASN_ENCODING or X509_ASN_ENCODING;
       hCertStore := CertOpenStore(prov,
                    ENCODING_TYPE,
                    0,
                    0,
                    @filename);

     end;
   4: begin  //Open a registry store
    //prov := CERT_STORE_PROV_SYSTEM_REGISTRY;
    flag := CERT_STORE_READONLY_FLAG;
    n := 'PKI-Test';
    hCertStore := CertOpenStore(prov,
                    0,
                    0,
                    flag,
                    pvPara);

     end;

end;
    //See if there was an error
    if hCertStore = 0 then
      begin
           lastErr := GetLastError;
           if lastErr = ERROR_INVALID_HANDLE then
              RaiseErr('Invalid Handle');
           if lastErr = ERROR_INVALID_PARAMETER then
              RaiseErr('Invalid Parameter')
           else RaiseErr('Error code: ' + inttostr(lastErr));
       end;
end;



function tCryptography.GetPropertyFromKPI(buff: array of char;
  prop: integer): string;
//This will extract the Properties from the
// CRYPT_KEY_PROV_INFO structure.
// 1= Container Name
// 2= CSP
// 3= Provider Type
// 4= Key Spec.
var
 CurPos, buflen, ProvId, i: integer;
 Container, CSPname, subStr: string;
 pWc, pWbs: pWideChar;
begin
     buflen := length(buff);
     pWbs := @buff;
     Result := '';
     CurPos := 28;
     Container := UnicodeToStr(@buff, CurPos);
     inc(CurPos, 2);
     CSPname := UnicodeToStr(@buff, curPos);
     pWc := pWbs+12;
     ProvId := integer(buff[12]);
     if Prop =1 then Result := Container;
     if Prop =2 then Result := CSPname;
     if Prop =3 then Result := intToStr(ProvId);
     if Prop =4 then Result := '?';
end;

function Tcryptography.UnicodeToStr;
//(pba: pByteArray, offset Start index): string
var
   i: integer;
begin
     Result := '';
     i := offSet;
     while pba[i]<>0 do
     begin
       Result := Result + chr(pba[i]);
       inc(i,2);
     end;
     offSet := i;
end;  //UnicodeToStr


procedure Tcryptography.CertClose;
var
   r: bool;
begin
    r := CertCloseStore(hCertStore,
                    CERT_CLOSE_STORE_FORCE_FLAG);
end;

//Decode a Base64 string into a byte array
procedure Tcryptography.B64ToArray(S: string; var buf: array of byte; var len: integer);
var
   i: integer;
   t: string;
begin
     if S = '' then exit;
     t := MimeDecodeString(S);
     len := length(t);
     //Strings count from 1
     for i := 1 to Len do
     begin
          buf[i-1] := byte(t[i]);
     end;
end;

//*****************************************************
//This is the main routine
//*****************************************************
function tCryptography.VerifyData;
var
   pbEncodedBlob, pbDecoded: array of byte;
   cbEncodedBlob, cbSignerCertInfo, cbDecoded: DWORD;
   CertInfoBlob: array of DWORD;
   size, i, nmsize: integer;
   lasterr: DWORD;
   HashStatus, ReturnStatus, DateValid: boolean;
   v: pointer;
   str, NameString: string;
   pTime: PFileTime;
//nested procedure
procedure SetReturn(msg: string; st: boolean);
begin
     ReturnStatus := st;
     ReturnStr := msg;
end;  //Nested procedure
//Nested procedure
procedure Cleanup;
begin
     if pCertContext <> nil then CertFreeCertificateContext( pCertContext);
     pCertContext := nil;
     if hCertStore <> 0 then CertCloseStore( hCertStore, CERT_CLOSE_STORE_FORCE_FLAG);
     hCertStore := 0;
     if hMsg <> nil then CryptMsgClose( hMsg);
     hMsg := nil;
end;  //Nested procedure

begin
     ReturnStatus := false;
     ReturnStr := '89802010';  //Start with error status
     FixSigTime(fdatesigned, datesigned); //Convert M DT to FileTime
     try     //Try Finally block
      try    //Try Except block
     //Lets build the Hash of the data
     HashStart;
     HashBuf( pbyte(fDataBuffer), Length(fDataBuffer));
     GetHashValue;
     HashEnd;

     //Use Base 64 decode
     cbEncodedBlob := MimeDecodedSize(length(fSignatureStr));
     setlength(pbEncodedBlob,cbEncodedBlob);
     size := cbEncodedBlob;
     //MimeDecode(Pchar(fSignatureStr)^, size, Pchar(pbEncodedBlob)^);
     B64ToArray(fSignatureStr, pbEncodedBlob, integer(cbEncodedBlob));

     //-------------------------------------------------
     //Build a starting structure
     hMsg := CryptMsgOpenToDecode(
            cENCODING_TYPE,        // Encoding type.
            0,                     // Flags.
            0,                     // Use the default message type.
                                   // The message type is
                                   // listed in the message header.
            0,                     // Cryptographic provider. Use NULL
                                   // for the default provider.
            Nil,                   // Recipient information.
            Nil);                  // Stream information.
     if hMsg = nil then Exit;

     //--------------------------------------------------------------------
     //  Update the message with an encoded blob.
     //  pbEncodedBlob, the encoded data,
     //  cbEnclodeBlob, the length of the encoded data,
     if(CryptMsgUpdate(
        hMsg,                 // Handle to the message
        pointer(pbEncodedBlob),        // Pointer to the encoded blob
        cbEncodedBlob,              // Size of the encoded blob
        TRUE))                // Last call
        then      ReturnStr := 'The encoded blob has been added to the message.'
        else
            RaiseErr('89802015^Digital signature corrupted.');

     //---------------
     // Get the number of bytes needed for a buffer
     // to hold the Decoded message.
     if CryptMsgGetParam(hMsg,                 //Handle to msg
                         CMSG_CONTENT_PARAM,   //Param type
                         0,                    //Index
                         nil,
                         @cbDecoded)
                         Then ReturnStr := 'The message param has been acquired'
                         else RaiseErr('Decode CMSG_CONTENT_PARAM failed');
     // Allocate memory
     SetLength(pbDecoded, cbDecoded);
     //Copy the content to the buffer
     if CryptMsgGetParam(hMsg,                 //Handle to msg
                         CMSG_CONTENT_PARAM,   //Param type
                         0,                    //Index
                         pbDecoded,            //Address for return data
                         @cbDecoded)           //Size of return data
                    Then ReturnStr := 'The message param has been acquired'
                    else RaiseErr('Decode CMSG_CONTENT_PARAM failed');
     //See if the Hash matches
     HashStatus := true;
     for i := 0 to cbDecoded - 1 do
          if pbDecoded[i] <> cHashValue[i] then HashStatus := false;


     if HashStatus = false then RaiseErr('89802016^Mismatched digital signature hash values.');

     //Verify the signature
     //First, Get the signer CERT_INFO from the message
     //------
     //Get the size needed
     if CryptMsgGetParam( hMsg,      //Msg Handle
                          CMSG_SIGNER_CERT_INFO_PARAM,   //Param Type
                          0,         //Index
                          nil,
                          @cbSignerCertInfo)  //Size of return data
        then ReturnStr:=IntToStr(cbSignerCertInfo) + ' bytes needed'
        else RaiseErr('Verify SIGNER_CERT #1 failed');
     //Allocate Memory
     SetLength(CertInfoBlob,cbSignerCertInfo);
     //Get the signer CERT_INFO
     if CryptMsgGetParam( hMsg,      //Msg Handle
                          CMSG_SIGNER_CERT_INFO_PARAM,   //Param Type
                          0,         //Index
                          Pointer(CertInfoBlob),
                          @cbSignerCertInfo)  //Size of return data
        then ReturnStr:='CertInfoBlob acquired'
        else RaiseErr('Verify SIGNER_CERT #2 failed');
     //------------
     //Open a certificate store in memory using CERT_STORE_PROV_MSG
     //which initializes it with the certificates from the MSG
     hCertStore := CertOpenStore( CERT_STORE_PROV_MSG,     //Store prov type
                                  cENCODING_TYPE,
                                  0,                       //Cryptographic provider
                                                           // use nil for default
                                  0,                       //Flags
                                  hMsg);
     if hCertStore = 0 then RaiseErr('Open Store failed');

     //--------------
     //Find the signer's cert in the store
     pCertContext :=
       CertGetSubjectCertificateFromStore(hCertStore,
                                          cENCODING_TYPE,
                                          pointer(CertInfoBlob));

     if pCertContext = nil then RaiseErr('GetSubjectCert failed');
     //Allocate memory for name
     nmsize := 256;
     SetLength(NameString, nmsize);
     //Get the Cert Name String
     CertGetNameString(pCertContext,
                       CERT_NAME_SIMPLE_DISPLAY_TYPE,
                       0,
                       0,
                       pointer(NameString),
                       nmsize );
     //---------
     //Use the CERT_INFO from the signer Cert to Verify
     //the signature
     if CryptMsgControl( hMsg,             //Handle to msg
                         0,                //Flags
                         CMSG_CTRL_VERIFY_SIGNATURE,
                         pCertContext.pCertInfo)   //Pointer to the CERT_INFO
                         then SetReturn('Verify signature succeeded', true)
                         else
                           begin
                           lasterr := getLastError; // Check last system error
                           SetReturn('89802018^Digital signature verification failed: ' +
                           SysErrorMessage(lasterr), false);
                           end;

     //Check that the time is valid
     pTime := nil;
     if fdatesigned <> '' then pTime := @datesigned;
     i := CertVerifyTimeValidity(
            pTime,                // Use time of signing or current time.
            pCertContext.pCertinfo);   // Pointer to CERT_INFO.
     case i of
     -1 : SetReturn('89802019^Before the certificate effective date.', false) ;   //Before the not before time
      1 : SetReturn('89802020^Certificate expired', false)  ;   //After the not after time
      0 : DateValid := true;
     end;

     v := pCertContext.pCertInfo;
     //We want to use the online certificate status protocol (OCSP)
     //so we wont have any CRL's
     {if ReturnStatus and (CheckCertAgenstCRL(v) = false) // (v)
        then  SetReturn('89802017^Certificate revoked.', false);
        }
     //Need to keep setting this
     Result := ReturnStatus;
     //-------------
     //Do a Cert Chain
     //CertChainCheck(pCertContext);
     if ReturnStatus and (CertChainCheck(pCertContext) = false) then
           SetReturn('Cert Chain Error: ' + ReturnStr, false);
     Result := ReturnStatus;
     //-------------
     //Check for revocation
     //CertRevocation(pCertContext);
     if ReturnStatus and (CertRevocation(pCertContext) = false) then
             SetReturn('Certificate Revoked', false);
     except
     if '8980' = copy(ErrCode, 1,4) then
             SetReturn( ErrCode, false)
     else SetReturn('89802010^Contact IRM: ' + ErrCode, false);
     end; //try except
     finally
     Cleanup;
     Result := ReturnStatus;
     end; //try finally
end;

//Should not need this because we are using OCSP checking.
function tCryptography.CheckCertAgenstCRL(v:pointer): boolean;  //(pcrtinfo: PCERT_INFO): boolean;
var
   //CRLInfo: CRLInfo;
   cnt: integer;
   pPrevCrlContext: PCCRL_CONTEXT;
   CRLcontextArray: array of PCRL_INFO; //Array[0..100] of Pointer;
begin
{O-}
     Result := true;
     SetLength(CRLcontextArray, 100);
     for cnt := 0 to 99 do  CRLcontextArray[cnt] := nil;

     try
     CRLOpenStore;
     cnt := 0;
     CRLcontextArray[0] :=nil;
     pPrevCrlContext := nil;
     pPrevCrlContext := CertEnumCRLsInStore( hCRLstore,
                                             nil);
     while (pPrevCrlContext <> nil) and (cnt < 100) do
     begin
     CRLcontextArray[cnt] := pPrevCrlContext.pCrlInfo;
     inc(cnt);
     CRLcontextArray[cnt] := nil;
     pPrevCrlContext := CertEnumCRLsInStore( hCRLstore,
                                             pPrevCrlContext);
     end;
     if cnt = 0 then Exit;  //Just quit
     if CertVerifyCRLRevocation( cENCODING_TYPE,
                                 v, //pCertContext.pCertInfo, //pcrtinfo,
                                 cnt,
                                 CRLcontextArray) = false
                                 then Result := false;

     finally
     CRLCloseStore;
     cnt :=0;
     end; //try
end;
//Open a Cert store with CRL's
// Lets use a FileName store
procedure tCryptography.CRLOpenStore;
var
   prov: LPCSTR;  //What cert store to use
   storeName: LPCTSTR;
   flag, lasterr: DWORD;  //Cert store control
   pvPara: PVOID; //Pointer to a string or nil
   lpcfn: LPCSTR;
   fn, n: string;
   c: integer;
begin
    hCRLStore := 0;
    fn := c_stoFilePath + c_stoFilename;
    c := 1; //Start with a memory store
    if FileExists( fn) then c := 3; //Use the File store

case c of
  1: begin  //Open a memory store
    prov := CERT_STORE_PROV_MEMORY;
    flag := CERT_SYSTEM_STORE_CURRENT_USER ;
    pvPara := nil;
    hCRLStore := CertOpenStore(prov,
                    0,
                    0,
                    0,
                    nil);
     end;  //1

  2: begin //Open a system store
       prov := CERT_STORE_PROV_SYSTEM_A;
       flag := CERT_SYSTEM_STORE_CURRENT_USER;
       hProv := 0;
       storeName := 'CRL';

       hCRLStore := CertOpenSystemStore( hprov,
                    storeName);
     end;  //2
  3: begin  //Open a Filename store
       lpcfn := pchar(fn);
       prov := CERT_STORE_PROV_FILENAME_A;  //Ascii name
       flag := CERT_SYSTEM_STORE_CURRENT_USER;
       hCRLStore := CertOpenStore(prov,
                    cENCODING_TYPE,
                    0,
                    0,
                    lpcfn);

     end;  //3
     end; //case

    //See if there was an error
    if hCRLStore = 0 then
      begin
           lastErr := GetLastError;
           if lastErr = ERROR_INVALID_HANDLE then
              RaiseErr('Invalid Handle');
           if lastErr = ERROR_INVALID_PARAMETER then
              RaiseErr('Invalid Parameter')
           else RaiseErr('Error code: ' + inttostr(lastErr));
       end;
end;

procedure tCryptography.CRLCloseStore;
begin
     if hCRLStore <> 0 then
                  CertCloseStore(hCRLStore,
                    CERT_CLOSE_STORE_FORCE_FLAG);
end;

function tCryptography.CertChainCheck(
  pCertContext: PCCERT_CONTEXT): boolean;
var
    { Private declarations }
   hChainEngine: HCERTCHAINENGINE;
   ChainConfig:  CERT_CHAIN_ENGINE_CONFIG;
   pChainContext: PCCERT_CHAIN_CONTEXT;
   pDupContext:  PCCERT_CHAIN_CONTEXT;
   pTime: PFILETIME;
   //hCertStore:   HCERTSTORE;
   StoreList: Array[0..2] of DWORD;
   EnhkeyUsage:  CERT_ENHKEY_USAGE;
   CertUsage:    CERT_USAGE_MATCH;
   ChainPara:    CERT_CHAIN_PARA;
   dwFlags:      DWORD;
   i: integer;
   fStatus: boolean;  //Local status
   //Msg: TStringList;
//-----------------------------------------------------
//nested function
procedure Memo(s: string);
begin
     TrackingMsg.Add(s);
end;

procedure SetStatus(sta: boolean; s: string);
begin
     fStatus := false;
     if sta then fStatus := True;
     if not sta then ReturnStr := s;
     Memo(s);
end;
//end nested function
//-------------------------------------------------------
begin
     pTime := nil;
     if fdatesigned <> '' then pTime := @datesigned;
     TrackingMsg.Clear;
     StoreList[0] := 0;
     StoreList[1] := 0;
     StoreList[2] := 0;
     CRLOpenStore;
     StoreList[0] := hCRLStore;
     //
     EnhkeyUsage.cUsageIdentifier := 0;
     EnhkeyUsage.rgpszUsageIdentifier := nil;

     CertUsage.dwType := USAGE_MATCH_TYPE_AND;
     CertUsage.Usage := EnhkeyUsage;
     //
     ChainPara.cbSize := sizeof(chainPara);
     ChainPara.RequestedUsage := CertUsage;
     //
     ChainConfig.cbSize := sizeof( CERT_CHAIN_ENGINE_CONFIG);
     ChainConfig.hRestrictedRoot := 0 ;
     ChainConfig.hRestrictedTrust := 0 ;
     ChainConfig.hRestrictedOther := 0 ;
     ChainConfig.cAdditionalStore :=1 ;            //0
     ChainConfig.rghAdditionalStore := DWORD(@StoreList) ; //0
     ChainConfig.dwFlags := //CERT_CHAIN_REVOCATION_CHECK_END_CERT;
                            // or CERT_CHAIN_REVOCATION_CHECK_CHAIN
                            //CERT_CHAIN_REVOCATION_CHECK_CACHE_ONLY;
                            CERT_CHAIN_CACHE_END_CERT;
     ChainConfig.dwUrlRetrievalTimeout := 3 ;
     ChainConfig.MaximumCachedCertificates :=0 ;
     ChainConfig.CycleDetectionModulus := 0;

     //---------------------------------------------------------
     //   Create the nondefault certificate chain engine.
     if CertCreateCertificateChainEngine(
                        @ChainConfig,
                        hChainEngine)
                        then Memo('A chain Engine has been created')
                        else RaiseErr('The engine Create Failed');
     //---------------------------------------------------------
     //----------------------------------------------------------------
     //        Build a chain using CertGetCertificateChain
     //        and the certificate retrieved.
     dwFlags := 0;
     dwFlags := //CERT_CHAIN_REVOCATION_CHECK_CACHE_ONLY;
                CERT_CHAIN_REVOCATION_CHECK_END_CERT;
                //or CERT_CHAIN_REVOCATION_CHECK_CHAIN_EXCLUDE_ROOT
                //or CERT_CHAIN_CACHE_END_CERT;
                //or CERT_CHAIN_RETURN_LOWER_QUALITY_CONTEXTS;

          if(CertGetCertificateChain(
                  hChainEngine,                  // Use 0 the default chain engine.
                  pCertContext,          // Pointer to the end certificate.
                  pTime,                  // Use the default time.
                  0,              // Search no additional stores.
                  @ChainPara,            // Use AND logic, and enhanced key usage
                                         // as indicated in the ChainPara
                                         // data structure.
                  dwFlags,
                  nil,                 // Currently reserved.
                  pChainContext))       // Return a pointer to the chain created.
                  then
                    Memo('The chain has been created. ')
                  else
                  RaiseErr('The chain could not be created.');

     //---------------------------------------------------------------
     // Display some of the contents of the chain.

   Memo('The size of the chain context is ' + IntToStr(pChainContext.cbSize));
   Memo(IntToStr(pChainContext.cChain) + ' simple chains found.');
   Memo('Error status for the chain:');
   CRLCloseStore;

   i :=  pChainContext.TrustStatus.dwErrorStatus;
   case i of
   CERT_TRUST_NO_ERROR :
     SetStatus(true, 'No error found for this certificate or chain.');

     CERT_TRUST_IS_NOT_TIME_VALID:
     SetStatus(false, 'This certificate or one of the certificates in the certificate chain is not time-valid.');

     CERT_TRUST_IS_NOT_TIME_NESTED:
     SetStatus(false, 'Certificates in the chain are not properly time-nested.');

     CERT_TRUST_IS_REVOKED:
     SetStatus(false, 'Trust for this certificate or one of the certificates in the certificate chain has been revoked.');

     CERT_TRUST_IS_NOT_SIGNATURE_VALID:
     SetStatus(false, 'The certificate or one of the certificates in the certificate chain does not have a valid signature.');

     CERT_TRUST_IS_NOT_VALID_FOR_USAGE:
     SetStatus(false, 'The certificate or certificate chain is not valid in its proposed usage.');

     CERT_TRUST_IS_UNTRUSTED_ROOT:
     SetStatus(false, 'The certificate or certificate chain is based on an untrusted root.');

     //***Check this Will allow this because all EPCS return this.
     CERT_TRUST_REVOCATION_STATUS_UNKNOWN:
     SetStatus(true, 'The revocation status of the certificate or one of the certificates in the certificate chain is unknown.');

     CERT_TRUST_IS_CYCLIC :
     SetStatus(false, 'One of the certificates in the chain was issued by a certification authority that the original certificate had certified.');

     CERT_TRUST_IS_PARTIAL_CHAIN:
     SetStatus(false, 'The certificate chain is not complete.');

     CERT_TRUST_CTL_IS_NOT_TIME_VALID:
     SetStatus(false, 'A CTL used to create this chain was not time-valid.');

     CERT_TRUST_CTL_IS_NOT_SIGNATURE_VALID:
     SetStatus(false, 'A CTL used to create this chain did not have a valid signature.');

     CERT_TRUST_CTL_IS_NOT_VALID_FOR_USAGE:
     SetStatus(false, 'A CTL used to create this chain is not valid for this usage.');
     end;  //case


     Memo('Info status for the chain:');
     case (pChainContext.TrustStatus.dwInfoStatus) of

     CERT_TRUST_HAS_EXACT_MATCH_ISSUER :
     Memo('An exact match issuer certificate has been found for this certificate.');

     CERT_TRUST_HAS_KEY_MATCH_ISSUER:
     Memo('A key match issuer certificate has been found for this certificate.');

     CERT_TRUST_HAS_NAME_MATCH_ISSUER:
     Memo('A name match issuer certificate has been found for this certificate.');

     CERT_TRUST_IS_SELF_SIGNED:
     Memo('This certificate is self-signed.');

     CERT_TRUST_IS_COMPLEX_CHAIN:
     Memo('The certificate chain created is a complex chain.');
     else  Memo('No information status reported.');
     end; // End case

     //--------------------------------------------------------------------
     Result := fStatus;
end;


function tCryptography.CertRevocation(
  pCertContext: PCCERT_CONTEXT): boolean;
var
   //var's for Revocation checking
   i: integer;
   str2: string;
   RevStatus: CERT_REVOCATION_STATUS;
   pRevStatus: PCERT_REVOCATION_STATUS;
   RevPara: CERT_REVOCATION_PARA;
   pRevPara: PCERT_REVOCATION_PARA;
   rgpvContext: array[0..3] of pointer;
   cContext: DWORD;
   RevocationStatus: BOOL;
   prgp: PPVOID;
begin
        //Check for Revocation
        RevocationStatus := true;  //Start with it true
        RevPara.cbSize := sizeof(RevPara);
        pRevPara := @RevPara; //Get a pointer to it
        RevStatus.cbSize := sizeof(RevStatus);
        RevStatus.dwIndex := 0;
        RevStatus.dwError := 0;
        pRevStatus := @RevStatus;
        for i := 0 to 3 do rgpvContext[i] := nil;

        rgpvContext[0] := pCertContext;
        cContext := 1;

        prgp := @(rgpvContext[0]);


        RevocationStatus :=
                CertVerifyRevocation( cENCODING_TYPE,
                               CERT_CONTEXT_REVOCATION_TYPE,
                               cContext,   //Count of elements in array.
                               prgp,       //Array of pointers to Cert_context
                               CERT_VERIFY_REV_SERVER_OCSP_FLAG,  // was CERT_VERIFY_REV_CHAIN_FLAG,
                               nil,
                               pRevStatus);
        //Now to check the status
        if RevocationStatus = false then
           begin
                LastErr := GetLastError;
                if LastErr < $80000000 then str2 := SysErrorMessage(LastErr);
                TrackingMsg.Add(SysErrorMessage(LastErr));
           end;  //if
         Result := RevocationStatus;

end;

procedure tCryptography.FixSigTime(MDT: string; var cryptoDT: filetime);
var
   str: string;
   dt: TdateTime;
   systime: TsystemTime;
   iyear: integer;
   day, month, year: string;
begin
     cryptoDT.dwLowDateTime := 0;
     cryptoDT.dwHighDateTime := 0;
     if fDatesigned = '' then exit;
     iyear := 1700 + StrToInt(copy(fDateSigned, 1, 3));
     year := IntToStr(iyear);
     month := copy(fDateSigned, 4, 2);
     day := copy(fDateSigned, 6, 2);
     if ShortdateFormat = 'M/d/yyyy' then
        str := month + DateSeparator + day + DateSeparator + year;
     if ShortDateFormat = 'd/M/yyyy' then
        str := day + DateSeparator + month + DateSeparator + year + '12:01';
     dt := StrToDate(str);
     DateTimeToSystemTime(dt, sysTime);
     SystemTimeToFileTime(sysTime, cryptoDT);

end;

end.


