unit fPKIserver;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, SvcMgr, Dialogs,
  IdThreadMgr, IdThreadMgrDefault, IdBaseComponent, IdComponent,
  IdTCPServer, IdCoder, {IdCoderMessageDigest,}
  //VA units
  XuDsigV, XlfFileInfo, XuDsigConst, IdThreadMgrPool;

type
  TPKIService = class(TService)
    IdTCPServer1: TIdTCPServer;
    IdThreadMgr1: TIdThreadMgrPool;
    procedure PKIServiceStart(Sender: TService; var Started: Boolean);
    procedure PKIServiceStop(Sender: TService; var Stopped: Boolean);
    procedure IdTCPServer1Connect(AThread: TIdPeerThread);
    procedure IdTCPServer1Execute(AThread: TIdPeerThread);
    function DoReadBuf(AThread: TIdPeerThread):string;
    procedure DoWriteStr(AThread: TIdPeerThread; Str: string);

  private
    { Private declarations }
  public
    function GetServiceController: TServiceController; override;
    { Public declarations }
  end;

var
  PKIService: TPKIService;
  CRCString: String;

implementation

{$R *.DFM}

procedure ServiceController(CtrlCode: DWord); stdcall;
begin
  PKIService.Controller(CtrlCode);
end;

function TPKIService.GetServiceController: TServiceController;
begin
  Result := ServiceController;
end;

procedure TPKIService.PKIServiceStart(Sender: TService;
  var Started: Boolean);
var
   crc: integer;
begin
     //The use of GetFileCrc32
     crc := XlfFileInfo.GetFileCrc32( paramstr(0) );
     CRCString := IntToHex(crc, 8);
     IdTCPServer1.DefaultPort := c_DsigPort;  //Get port from XuDsigConst
     IdTCPServer1.Active := true;
     Started := true;
end;

procedure TPKIService.PKIServiceStop(Sender: TService;
  var Stopped: Boolean);
begin
     IdTCPServer1.Active := false;
     {ask all threads to stop}
//     IdThreadMgr1.TerminateThreads(IdThreadMgr1.ActiveThreads);
     IdThreadMgr1.TerminateThreads();
     {Should wait to see that all threads have stopped first}
     Stopped := True;
end;



procedure TPKIService.IdTCPServer1Connect(AThread: TIdPeerThread);
var
   str: string;
begin
     //ConnCnt := ConnCnt + 1;

     with AThread.Connection do
     begin
          Writeln('230 Welcome to the PKI Verify Server');
          Writeln('230 Server CRC: ' + CRCstring);
          //Writeln('QUIT');
          //Writeln('Any other strings will be returned "as is"');
      end;
end;

procedure TPKIService.IdTCPServer1Execute(AThread: TIdPeerThread);
{This is the main work loop for the simple SMTP like protocol.
 The commands are HELO, DATA, TURN, QUIT.  If a command is understood
 then a 220 is returned else 440 is returned.
 The real work is done in the DATA command.
 The VistA system sends a TURN command to tell us it is ready
 for us to send the responce data, (what ever it is)
 Need to add some secutity to the HELO
}
var
   Msg, Cmd, Ans: string;
   i, Cnt, ServiceCase: integer;

begin
     Cnt := 0;
     while AThread.Connection.Connected do
     begin
        Msg := AThread.Connection.ReadLn;  //Read to a CR/LF
        Cmd := '\' + UpperCase(Copy(Msg,1,4)) + '\';
        i := Pos(Cmd, '\HELO\QUIT\SERV\DATA\TURN\NOOP\');
        case i of
          1: //HELO
             AThread.Connection.Writeln('220 ');

          6: //QUIT     Close the connection
             begin
               AThread.Connection.Writeln('220 Quit');
               AThread.Connection.Disconnect;
             end;

          11: //Service   Not used
              begin
                //The services are converted to a number on the M side
                // 1= ECHO,
                ServiceCase := StrtoInt(Copy(Msg, 6, (length(Msg)-5)));
              end;

          16: //Data     Read the data, This may be where
              begin
                Ans := DoReadBuf(AThread);
                AThread.Connection.Writeln('220 Data');
              end;

          21: //Turn
              begin
                   AThread.Connection.WriteLn('220 TURN');
                   AThread.Connection.WriteLn('DATA ');
                   DoWriteStr(AThread, Ans);
                   Msg := AThread.Connection.Readln;
              end;

          26: //NOOP
              AThread.Connection.WriteLn('220 NOOP');
          else
          begin
              //Cmd's not understood
              AThread.Connection.WriteLn('440 Unknown "' + Msg + '"');
          end;

     end; //Case
  end; { AThread.Connection.Connected}
end;

function TPKIService.DoReadBuf(AThread: TIdPeerThread):string;
//Read in the databuffer and then the Digital signature.
//Call the PKI verify code and return the answer.
//Eack data block will end with a zero length record.
//The data that was signed is the first block.
//The second block is the signature.

//read the data format 3digit length for the data that follows.
// a length of -10 is end of data

var
  Str,S, DataString, SigString, SignDate: string;
  Len, i, block: integer;
  Crypto: Tcryptography;
  Buf: Array[0..255] of byte;
  CryptoComment: String;
  FileName: String;
  DirName: String;
  MemoList: TStringList;
begin
   Len:=0;
   block := 0;
   DataString :='';
   SigString := '';
   SignDate := '';
   //Create the Crypto object first
   Crypto := tCryptography.Create;
   CryptoComment := Crypto.Comment;
   //Now get the data

  while Len >= 0 do
  begin
    S := AThread.Connection.ReadString(3);
    Str := '';
    Len := StrtoInt(S);
    if Len > 0 then //if length is zero nothing to read
    begin
      //Read the data into the string
      Str := AThread.connection.ReadString(Len);
      Case block of
        0: DataString := DataString + Str;
        1: SigString := SigString + Str;
        2: SignDate := SignDate + Str;
      end; //case

    end; //Begin
    if Len = 0 then Inc(block);
  end;
   //Now build crypto and check
  try
    Crypto.Databuffer := DataString;
    Crypto.SignatureStr := SigString;
    Crypto.DTSigned := SignDate;
    if Crypto.VerifyData then
      Result := 'OK'
    else
    begin
      Result := '-1^' + crypto.ReturnStr;
      Result := '-1^'+Crypto.ReturnStr;
      DirName := 'C:\PKIServer';
      FileName := DirName +FormatDateTime('"\ePCSVerifyError_"yyyymmdd"@"hhmmss".txt"',Now);
      MemoList := TStringList.Create;
      MemoList.Add('Date Signed: ' + Crypto.DTSigned);
      MemoList.Add('Returned Msg: ' + Crypto.ReturnStr);
      MemoList.Add('Tracking Data:');
      //Memo1.Lines.AddStrings(Crypto.TrackingMsg);
      for i := 0 to Crypto.TrackingMsg.Count-1 do
      begin
        MemoList.Add('  '+Crypto.TrackingMsg.Strings[i]);
      end;
      if not DirectoryExists(DirName) then
        if not CreateDir(DirName) then
          raise Exception.Create('Cannot create '+DirName);
      MemoList.SaveToFile(FileName);
      MemoList.Free;
      MemoList := nil;
    end;
  except
    on E: eCryptErr do
      Result := '-2^Error: ' + E.Message +
       'Initial Comment: '+CryptoComment+'  Current Comment: '+Crypto.Comment ;
  end;
  Crypto.destroy;  //Done with it.
end;

procedure TPKIService.doWriteStr(AThread: TIdPeerThread; Str: string);
//Write the data from
var
   S: string;
   Len: integer;
begin
     //Str := value;
     Len := Length(Str);
     S := Copy(IntToStr(Len+1000),2,3);
     AThread.Connection.Write(S);
     if Len > 0 then
        AThread.Connection.Write(Str);
     //End of DATA

     AThread.Connection.Write('-10');
end;

end.
