unit fDataEntryForPrescriber;

interface

uses
  System.Variants,
  System.Classes,
  System.SysUtils,
  System.Actions,
  System.UITypes,
  Vcl.Graphics,
  Vcl.Controls,
  Vcl.Forms,
  Vcl.Dialogs,
  Vcl.StdCtrls,
  Vcl.ExtCtrls,
  Vcl.ComCtrls,
  Vcl.Menus,
  Vcl.ActnList,
  Winapi.Messages,
  Windows,
  Diaccess,
  Fmcntrls,
  Fmcmpnts,
  Fmlookup,
  Trpcb,
  oPCSPrescriber;

type
  TfrmPrescriberDateEntryForm = class(TForm)
    pnlTop: TPanel;
    edtServer: TEdit;
    lblServer: TLabel;
    edtPort: TEdit;
    lblPortNumber: TLabel;
    btnGetServerInfo: TButton;
    btnConnect: TButton;
    fmlkupPrescriber: TFMLookUp;
    FMFinder1: TFMFinder;
    FMLister1: TFMLister;
    RPCBroker1: TRPCBroker;
    mmMain: TMainMenu;
    File1: TMenuItem;
    Exit1: TMenuItem;
    Help1: TMenuItem;
    About1: TMenuItem;
    fmGetsPrescriber: TFMGets;
    fmGetsActiveUser: TFMGets;
    FMFiler1: TFMFiler;
    ActionList1: TActionList;
    btnExit: TButton;
    FMValidator1: TFMValidator;
    timAllSchedules: TTimer;
    FMLister2: TFMLister;
    acConnectDisconnect: TAction;
    pnlMain: TPanel;
    lblDeaExpirationDate: TLabel;
    lblDEANumber: TLabel;
    lblDetoxMaintNumber: TLabel;
    lblVANumber: TLabel;
    FMLabel1: TFMLabel;
    Label2: TLabel;
    Label5: TLabel;
    bvl: TBevel;
    lblScheduleHeader: TLabel;
    Label1: TLabel;
    btnUpdate: TButton;
    fmedtVANumber: TFMEdit;
    fmedtProviderName: TFMEdit;
    fmedtDetoxMaintNumber: TFMEdit;
    fmedtDeaExpirationDate: TFMEdit;
    fmedtDeaNumber: TFMEdit;
    fmcbxScheduleV: TFMCheckBox;
    fmcbxScheduleIV: TFMCheckBox;
    fmcbxScheduleIINonNarcotic: TFMCheckBox;
    fmcbxScheduleIINarcotic: TFMCheckBox;
    fmcbxScheduleIIINonNarcotic: TFMCheckBox;
    fmcbxScheduleIIINarcotic: TFMCheckBox;
    fmcbxAuthorizedToWriteOrders: TFMCheckBox;
    btnSelectPrescriber: TButton;
    fmedtAltName: TFMEdit;
    fmedtAccessCode: TFMEdit;
    fmedtVerifyCode: TFMEdit;
    fmedtDISUSERed: TFMEdit;
    fmedtTerminationDate: TFMEdit;
    mmoUtility: TMemo;
    edtFacilityDEANum: TEdit;
    Label3: TLabel;
    Label4: TLabel;
    Label6: TLabel;
    acAbout: TAction;
    acGetServerInfo: TAction;
    acExit: TAction;
    acSelectPrescriber: TAction;
    acClearMainPanel: TAction;
    Bevel1: TBevel;
    acUpdate: TAction;
    procedure FormCreate(Sender: TObject);

    procedure acConnectDisconnectExecute(Sender: TObject);
    procedure acAboutExecute(Sender: TObject);
    procedure acGetServerInfoExecute(Sender: TObject);
    procedure acExitExecute(Sender: TObject);
    procedure acSelectPrescriberExecute(Sender: TObject);
    procedure acClearMainPanelExecute(Sender: TObject);
    procedure fmedtAltNameExit(Sender: TObject);
    procedure FormCloseQuery(Sender: TObject; var CanClose: Boolean);
    procedure acUpdateExecute(Sender: TObject);

    private
      FCurrentPrescriber: TPCSPrescriber;

      function FMDateTime2TDateTime(aValue: string): TDateTime;
      function IsActiveUser: Boolean;
      function HoldsVistAKey(AKeyName: string): Boolean;
      function CheckSchedules: Boolean;
      function GetFacilityVANum(IENS: string): string;
      function GetValueFromVistA(InputStr: string): string;
      function GetSANFromVistA: string;

      procedure InitializeData(aPCSPrescriber: TPCSPrescriber);
  end;

var
  frmPrescriberDateEntryForm: TfrmPrescriberDateEntryForm;

implementation

uses
  fWarning,
  fVistAAbout,
  XuDSigS,
  RpcConf1,
  MFunStr;

{$R *.dfm}

procedure TfrmPrescriberDateEntryForm.FormCreate(Sender: TObject);
var
  i: integer;
begin
  for i := 0 to ParamCount do
    if (Pos('P=', UpperCase(ParamStr(i))) = 1) then
      edtPort.Text := Copy(ParamStr(i), 3, Length(ParamStr(i)))
    else if (Pos('S=', UpperCase(ParamStr(i))) = 1) then
      edtServer.Text := Copy(ParamStr(i), 3, Length(ParamStr(i)));

  Self.ShowHint := True;
  Application.ShowHint := True;
  pnlTop.DoubleBuffered := True;
  pnlMain.DoubleBuffered := True;
  acClearMainPanel.Execute;
end;

procedure TfrmPrescriberDateEntryForm.FormCloseQuery(Sender: TObject; var CanClose: Boolean);
begin
  FreeAndNil(FCurrentPrescriber);
  CanClose := True;
end;

procedure TfrmPrescriberDateEntryForm.acAboutExecute(Sender: TObject);
begin
  frmVistAAbout.Execute;
end;

procedure TfrmPrescriberDateEntryForm.acClearMainPanelExecute(Sender: TObject);
var
  i: integer;
  CurrObject: TControl;
begin
  for i := 0 to pnlMain.ControlCount - 1 do
    begin
      CurrObject := pnlMain.Controls[i];
      if CurrObject is TFMCheckBox then
        TFMCheckBox(CurrObject).Checked := False
      else if CurrObject is TFMEdit then
        TFMEdit(CurrObject).Text := '';

      { If no prescriber or Broker connection, disable the control as well }
      if CurrObject is TControl then
        TControl(CurrObject).Enabled := (Assigned(FCurrentPrescriber) and RPCBroker1.Connected);
    end;

  { Always gets cleared manually here, not FM Components }
  edtFacilityDEANum.Text := '';

  { Enabled depends on Broker Status }
  acSelectPrescriber.Enabled := RPCBroker1.Connected;
  btnSelectPrescriber.Enabled := RPCBroker1.Connected;
end;

procedure TfrmPrescriberDateEntryForm.acConnectDisconnectExecute(Sender: TObject);
var
  aCryptography: TCryptography;
  aVistaSubjAltName: string;
  aSubjAltName: string;
begin
  if RPCBroker1.Connected then
    begin
      RPCBroker1.Connected := False;
      FreeAndNil(FCurrentPrescriber);
      acClearMainPanel.Execute;
      acConnectDisconnect.Caption := '&Connect';
      Exit;
    end;

  if ((edtServer.Text = '') or (edtPort.Text = '')) then
    begin
      MessageDlg('Please selected a Server and Port combination before trying to Connect.', MtInformation, [mbOk], 0);
      Exit;
    end;

  try
    RPCBroker1.Server := ShortString(edtServer.Text);
    RPCBroker1.ListenerPort := StrToInt(edtPort.Text);

    RPCBroker1.ClearParameters := True;
    RPCBroker1.Connected := True;

    if (not RPCBroker1.Connected) then
      raise Exception.Create('Connection not established');

    if not RPCBroker1.CreateContext('XU EPCS EDIT DATA') then
      raise Exception.Create('User must have the XU EPCS EDIT DATA context option in their menu tree - could not be created!');

    if HoldsVistAKey('ORES') then
      raise Exception.Create('You cannot use this application if you hold the ORES key.');

    if (not HoldsVistAKey('XUEPCSEDIT')) then
      raise Exception.Create('You must hold the XUEPCSEDIT key to use this application.');

    aVistaSubjAltName := GetSANFromVistA;

    aCryptography := TCryptography.Create();
    try
      aCryptography.VistaUserName := RPCBroker1.User.Name;
      aSubjAltName := GetSANFromCard(Self, aCryptography);
    finally
      aCryptography.Free;
    end;

    if aSubjAltName = '' then
      raise Exception.Create('Invalid PIN or entry cancelled.');

    { We have a valid SubjAltName and the VistA file 200, field 501.2 is blank - Let's just update that while we're here }
    if aVistaSubjAltName = '' then
      with RPCBroker1 do
        try
          RemoteProcedure := 'XUS PKI SET UPN';
          Param[0].PType := Literal;
          Param[0].Value := aSubjAltName;
          Call;
        except
          raise;
        end;

    if CompareText(aVistaSubjAltName, aSubjAltName) <> 0 then
      raise Exception.Create('Subject Alternate Name on card and in VistA do not Match.');

    { We are connected and ready to run the app }
    acConnectDisconnect.Caption := '&Disconnect';
    FMValidator1.IENS := '';
    FreeAndNil(FCurrentPrescriber);
    acClearMainPanel.Execute;
  except
    on E: Exception do
      begin
        MessageDlg('Error: ' + E.Message, mtError, [mbOk], 0);
        FMValidator1.IENS := '';
        FreeAndNil(FCurrentPrescriber);
        acClearMainPanel.Execute;
        RPCBroker1.Connected := False;
        acConnectDisconnect.Caption := '&Connect';
      end;
  end;
end;

procedure TfrmPrescriberDateEntryForm.acExitExecute(Sender: TObject);
begin
  Close;
end;

procedure TfrmPrescriberDateEntryForm.acGetServerInfoExecute(Sender: TObject);
var
  aServer: string;
  aPort: string;
begin
  if GetServerInfo(aServer, aPort) <> MrCancel then
    begin
      edtServer.Text := aServer;
      edtPort.Text := aPort;
    end;
end;

procedure TfrmPrescriberDateEntryForm.acSelectPrescriberExecute(Sender: TObject);
var
  AddRecord: Boolean;
begin
  FreeAndNil(FCurrentPrescriber);
  FMValidator1.IENS := '';
  acClearMainPanel.Execute;
  fmlkupPrescriber.AllowNew := False;
  if fmlkupPrescriber.Execute(AddRecord) then
    begin
      FCurrentPrescriber := TPCSPrescriber.Create(fmlkupPrescriber.RecordNumber);

      fmGetsActiveUser.IENS := FCurrentPrescriber.IENS;
      fmGetsActiveUser.GetandFill;
      if IsActiveUser then
        begin
          acClearMainPanel.Execute;
          InitializeData(FCurrentPrescriber);
          FMValidator1.IENS := FCurrentPrescriber.IENS;
        end
      else
        begin
          FreeAndNil(FCurrentPrescriber);
          FMValidator1.IENS := '';
        end;
    end
  else
    MessageDlg('No record selected.', MtInformation, [mbOk], 0);
end;

procedure TfrmPrescriberDateEntryForm.acUpdateExecute(Sender: TObject);
var
  i: integer;
  changeList: TStringList;

  procedure AddTextChange(fmedt: TFMEdit; aOriginalValue: string);
  begin
    if (AnsiCompareStr(fmedt.FMCtrlInternal, aOriginalValue) <> 0) then
      changeList.Add(Format('`%s^`%s^%s^%s^%s',
        [FCurrentPrescriber.IEN, RPCBroker1.User.DUZ, fmedt.FMField, aOriginalValue, fmedt.FMCtrlInternal]));
  end;

  procedure AddBooleanChange(fmcbx: TFMCheckBox; aOriginalValue: Boolean);
  begin
    if not(fmcbx.Checked = aOriginalValue) then
      changeList.Add(Format('`%s^`%s^%s^%s^%s',
        [FCurrentPrescriber.IEN, RPCBroker1.User.DUZ, fmcbx.FMField, BoolToStr(aOriginalValue, True), BoolToStr(fmcbx.Checked, True)]));
  end;

begin
  if not CheckSchedules then
    begin
      MessageDlg('Update was Cancelled.', MtInformation, [mbOk], 0);
      Exit;
    end;

  { Build list of before and after for changed values }
  try
    changeList := TStringList.Create;
    AddTextChange(fmedtAltName, FCurrentPrescriber.SubjectAlternateName);
    AddTextChange(fmedtDeaNumber, FCurrentPrescriber.DEA);
    AddTextChange(fmedtVANumber, FCurrentPrescriber.VANumber);
    AddTextChange(fmedtDetoxMaintNumber, FCurrentPrescriber.DetoxNumber);
    AddTextChange(fmedtDeaExpirationDate, FCurrentPrescriber.DEAExpirationDate);

    AddBooleanChange(fmcbxAuthorizedToWriteOrders, FCurrentPrescriber.AuthorizedToWrite);
    AddBooleanChange(fmcbxScheduleV, FCurrentPrescriber.AllowScheduleVNarc);
    AddBooleanChange(fmcbxScheduleIV, FCurrentPrescriber.AllowScheduleIVNarc);
    AddBooleanChange(fmcbxScheduleIINonNarcotic, FCurrentPrescriber.AllowScheduleIINonNarc);
    AddBooleanChange(fmcbxScheduleIINarcotic, FCurrentPrescriber.AllowScheduleIINarc);
    AddBooleanChange(fmcbxScheduleIIINonNarcotic, FCurrentPrescriber.AllowScheduleIIINonNarc);
    AddBooleanChange(fmcbxScheduleIIINarcotic, FCurrentPrescriber.AllowScheduleIIINarc);

    if changeList.Count > 0 then
      try
        FMFiler1.Update;

        with RPCBroker1
          do
          begin
            RemoteProcedure := 'XU EPCS EDIT';
            Param[0].PType := list;
            for i := 0 to changeList.Count - 1 do
              Param[0].Mult[IntToStr(i)] := changeList[i];
            Call;
          end;

        { Get a fresh copy of the data from the components }
        InitializeData(FCurrentPrescriber);
        MessageDlg('Update Completed', MtInformation, [mbOk], 0);
      except
        on E: Exception do
          MessageDlg('Error updating record: ' + E.Message, mtError, [mbOk], 0);
      end
    else
      MessageDlg('No updates found', MtInformation, [mbOk], 0);
  finally
    FreeAndNil(changeList);
  end;
end;

function TfrmPrescriberDateEntryForm.HoldsVistAKey(AKeyName: string): Boolean;
begin
  with RPCBroker1 do
    try
      RemoteProcedure := 'XUS KEY CHECK';
      Param[0].Value := AKeyName;
      Param[0].PType := Literal;
      Call;
      Result := (AnsiCompareText(Results[0], '1') = 0);
    except
      Result := False;
    end;
end;

function TfrmPrescriberDateEntryForm.GetSANFromVistA: string;
begin
  try
    RPCBroker1.RemoteProcedure := 'XUS PKI GET UPN';
    RPCBroker1.Call;
    if RPCBroker1.Results.Count > 0 then
      Result := RPCBroker1.Results[0];
  except
    Result := '';
  end;
end;

function TfrmPrescriberDateEntryForm.IsActiveUser: Boolean;
var
  aTerminationDate: TDateTime;
begin
  with TStringList.Create do
    try
      if fmedtAccessCode.Text = '' then
        Add('  - has NO ACCESS CODE');

      if fmedtVerifyCode.Text = '' then
        Add('  - has NO VERIFY CODE');

      if fmedtDISUSERed.Text = 'YES' then
        Add('  - is DISUSERed');

      if fmedtTerminationDate.FMCtrlInternal <> '' then
        begin
          aTerminationDate := FMDateTime2TDateTime(fmedtTerminationDate.FMCtrlInternal);
          if (Now >= aTerminationDate) then
            Add(Format('  - was terminated as of %s', [FormatDateTime('MMM D, YYYY', aTerminationDate)]));
        end;

      if Count > 0 then
        begin
          Insert(0, 'This is NOT an active prescriber.');
          Insert(1, '');
          Insert(2, 'This prescriber:');
          Add('');
          Add('Press Yes to continue processing with inactive prescriber. Press No to select a different prescriber.');
          Result := MessageDlg(Text, MtConfirmation, [mbYes, mbNo], 0) = mrYes;
        end
      else
        Result := True;
    finally
      Free;
    end;
end;

function TfrmPrescriberDateEntryForm.FMDateTime2TDateTime(aValue: string): TDateTime;
var
  y, m, d: Word;
begin
  y := (1700 + StrToInt(Copy(aValue, 1, 3)));
  m := StrToInt(Copy(aValue, 4, 2));
  d := StrToInt(Copy(aValue, 6, 2));
  Result := EncodeDate(y, m, d) + EncodeTime(0, 0, 0, 0);
end;

procedure TfrmPrescriberDateEntryForm.fmedtAltNameExit(Sender: TObject);
var
  Mesg, PartStr, Name: string;
  AtPos: integer;
  Bad: Boolean;
begin
  // Regular expression can be used to validate if an email address is in correct format.
  // For example: you can use this "^[A-Z0-9._%+-]+@[A-Z0-9.-]+\.[A-Z]{2,6}$" pattern to verify if the email address has valid format.

  Bad := False;
  // fmedtAltName.Text := LowerCase(fmedtAltName.Text);
  name := fmedtAltName.Text;
  AtPos := Pos('@', name);
  if not(AtPos > 0) then
    begin
      Bad := True;
    end
  else if not(Pos('.', fmedtAltName.Text) < AtPos) then
    begin
      Bad := True;
    end
  else
    begin
      PartStr := Copy(fmedtAltName.Text, AtPos + 1, Length(fmedtAltName.Text));
      if not(Pos('.', PartStr) > 1) then
        begin
          Bad := True;
        end;
    end;
  if Bad then
    begin
      Mesg := 'This is the user''s primary work e-mail address' + #10#13;
      Mesg := Mesg + 'It must be in the format' + #10#13#10#13;
      Mesg := Mesg + 'first.last@agency.type,' + #10#13#10#13;
      Mesg := Mesg + '(e.g., john.public@domain.ext)' + #10#13#10#13#10#13;
      Mesg := Mesg + 'where first.last are first and last names' + #10#13;
      Mesg := Mesg + '(and the last name may followed by a numeric digit).';
      ShowMessage(Mesg);
    end;
end;

procedure TfrmPrescriberDateEntryForm.InitializeData(aPCSPrescriber: TPCSPrescriber);
var
  i: integer;
  CurrObject: TObject;
begin
  edtFacilityDEANum.Text := GetFacilityVANum(aPCSPrescriber.IENS);
  fmGetsPrescriber.IENS := aPCSPrescriber.IENS;
  fmGetsPrescriber.GetandFill;

  { Why is this needed? The checkbox can't process a boolean??? }
  for i := 0 to pnlMain.ControlCount - 1 do
    begin
      CurrObject := pnlMain.Controls[i];
      if CurrObject is TFMCheckBox then
        TFMCheckBox(CurrObject).Checked := (TFMCheckBox(CurrObject).FMCtrlInternal = '1');
    end;

  with FCurrentPrescriber do
    begin
      AuthorizedToWrite := fmcbxAuthorizedToWriteOrders.Checked;
      SubjectAlternateName := fmedtAltName.FMCtrlInternal;
      DEA := fmedtDeaNumber.FMCtrlInternal;
      VANumber := fmedtVANumber.FMCtrlInternal;
      DetoxNumber := fmedtDetoxMaintNumber.FMCtrlInternal;
      DEAExpirationDate := fmedtDeaExpirationDate.FMCtrlInternal;

      { A blank from fileman is used as False }
      AllowScheduleIINarc := fmcbxScheduleIINarcotic.Checked;
      AllowScheduleIINonNarc := fmcbxScheduleIINonNarcotic.Checked;
      AllowScheduleIIINarc := fmcbxScheduleIIINarcotic.Checked;
      AllowScheduleIIINonNarc := fmcbxScheduleIIINonNarcotic.Checked;
      AllowScheduleIVNarc := fmcbxScheduleIV.Checked;
      AllowScheduleVNarc := fmcbxScheduleV.Checked;
    end;
end;

{ *
  * This function checks for whether any of the Drug Schedules
  * are checked.  If none of them are checked, a warning is
  * displayed indicating that the Prescriber's DEA number, etc.\
  * would be removed the next time they try to prescribe a
  * controlled substance if none of the schedules are checked.
  * The user can continue to save the changes or cancel
  *
  * Return true if one or more of the Drug Schedules are selected
  * Return true if none selected and Continue is selected to continue
  *        updating the VistA server
  * Return false if none selected and Cancel is selected to discontinue
  *        updating the VistA server
  * }
function TfrmPrescriberDateEntryForm.CheckSchedules: Boolean;
begin
  if (not fmcbxScheduleIINonNarcotic.Checked) and
    (not fmcbxScheduleIINarcotic.Checked) and
    (not fmcbxScheduleIIINonNarcotic.Checked) and
    (not fmcbxScheduleIIINarcotic.Checked) and
    (not fmcbxScheduleIV.Checked) and
    (not fmcbxScheduleV.Checked) then
    Result := (frmWarning.Execute = mrYes)
  else
    Result := True;
end;

function TfrmPrescriberDateEntryForm.GetFacilityVANum(IENS: string): string;
var
  Ival: integer;
  FacilityIen: string;
begin
  Result := '';
  mmoUtility.Clear;
  FMLister1.IENS := ',' + IENS;

  FMLister1.GetList(mmoUtility.Lines);
  FacilityIen := '';
  for Ival := 1 to mmoUtility.Lines.Count do
    begin
      if Pos(' 1', mmoUtility.Lines[Ival - 1]) > 0 then
        begin
          FacilityIen := mmoUtility.Lines[Ival - 1];
          FacilityIen := Copy(FacilityIen, 1, Length(FacilityIen) - 2);
        end;
    end;
  if (FacilityIen = '') and (mmoUtility.Lines.Count = 1) then
    begin
      FacilityIen := mmoUtility.Lines[0];
      if Pos(' 0', FacilityIen) > 0 then
        FacilityIen := Copy(FacilityIen, 1, Length(FacilityIen) - 2);
    end;
  if not(FacilityIen = '') then
    begin
      while Copy(FacilityIen, Length(FacilityIen), 1) = ' ' do
        FacilityIen := Copy(FacilityIen, 1, Length(FacilityIen) - 1);
      Result := Piece(GetValueFromVistA('$G(^DIC(4,' + FacilityIen + ',"DEA"))'), '^');
      if not(Result = '') then
        Exit;
    end;
  // No Division or Couldn't get Facility DEA Number for Division
  FacilityIen := GetValueFromVistA('$G(junk,$$SITE^VASITE())');
  FacilityIen := Piece(FacilityIen, '^');
  if not(FacilityIen = '') then
    begin
      Result := Piece(GetValueFromVistA('$G(^DIC(4,' + FacilityIen + ',"DEA"))'), '^')
    end;
end;

function TfrmPrescriberDateEntryForm.GetValueFromVistA(InputStr: string): string;
begin
  with RPCBroker1 do
    try
      RemoteProcedure := 'XWB GET VARIABLE VALUE';
      Param[0].Value := InputStr;
      Param[0].PType := Reference;
      Call;
      if (Results.Count > 0) then
        Result := Results[0]
      else
        Result := '';
    except
      Result := '';
    end;
end;

end.
